/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2000-2008 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, Inc.,
 *   59 Temple Place Ste 330, Boston MA 02111-1307, USA, version 2.1,
 *   incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/*
 * testrecovery.c
 *
 * This test verifies that we can indeed recover from a failure gracefully.
 * It is basically the same as the allocation test, except it keeps data-
 * structures in persistent memory and periodically kills off the processes.
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "lpsm.h"

#define COUNT	262144
#define SLOTS   4096
#define MAXLOG  26

void **areas;
int *sizes;

void verify_no_overlap(int slot)
{
  char *start = areas[slot];
  char *end = start+sizes[slot];
  char *xs, *xe;
  int i;

  for ( i = 0 ; i < SLOTS ; i++ ) {
    if ( i != slot && areas[i] ) {
      xs = areas[i];
      xe = xs + sizes[i];

      if ( xs < end && xe > start ) {
	printf("OVERLAP ALERT: [%p,%p) -> [%p,%p)\n",
	       start, end, xs, xe);
	abort();
      }
    }
  }
}

struct misc_info {
  long occupied;
  int count;
};

int main(int argc, char *argv[])
{
  int i, slot, newsize;
  double rnd;
  void **root_ptr;
  struct misc_info *misc_ptr;
  void *nptr;
  pid_t pid;
  pid_t master_pid = getpid();

  /* Do initial setup */

  setlinebuf(stdout);
  setlinebuf(stderr);
  
  unlink("arena.dat");
  unlink("arena.log");

  if ( !(root_ptr = lpsm_init("arena.dat", "arena.log")) ) {
    fprintf(stderr, "%s: lpsm_arena_init() failed\n", argv[0]);
    return 1;
  }

  printf("Arena:   initialized, %d root pointers at %p\n",
	 LPSM_ROOT_POINTERS, root_ptr);

  root_ptr[0] = misc_ptr = lpsm_malloc(sizeof(struct misc_info));
  root_ptr[1] = areas = lpsm_malloc(sizeof(void *) * SLOTS);
  root_ptr[2] = sizes = lpsm_malloc(sizeof(int) * SLOTS);

  misc_ptr->count = misc_ptr->occupied = 0;

  memset(areas, 0, sizeof(areas));
  memset(sizes, 0, sizeof(sizes));

  lpsm_checkpoint(0.0, PSMSYNC_SYNC);
  lpsm_shutdown();

  /* Become our own process group and ignore SIGTERM to make it easier to
     "kill everything".  Become our own session to keep shell job control
     from interfering with stop signals. */

  setpgrp();

  while ( 1 ) {
    int master_waits = rand() & 1;

    pid = fork();
    if ( pid < 0 ) {
      perror("fork");
      return 1;
    } else if ( pid > 0 ) {
      int status;

      /* Give the child process a reasonable chance to finish log replay,
	 otherwise we'll never get through the test.  On the other hand,
	 we want to check for failures during log replay too! */
      if ( master_waits ) {
	waitpid(pid, &status, WUNTRACED);
	if ( !WIFSTOPPED(status) ) {
	  printf("*** CHILD ERROR, exit code = 0x%x\n", status);
	  exit(1);
	}
	kill(pid, SIGCONT);
      }
      sleep(rand() & 63);
      signal(SIGTERM, SIG_IGN);
      kill(-master_pid, SIGTERM);
      signal(SIGTERM, SIG_DFL);
      waitpid(pid, &status, 0);

      if ( WIFEXITED(status) )
	break;			/* Child exited, not killed, consider us done */
    } else {
      struct timeval start_time, init_time;

      pid = getpid();
      printf("*** Process starting: %d, master_waits = %d\n",
	     pid, master_waits);

      gettimeofday(&start_time, NULL);

      root_ptr = lpsm_init("arena.dat", "arena.log");
      
      gettimeofday(&init_time, NULL);

      misc_ptr = root_ptr[0];
      areas    = root_ptr[1];
      sizes    = root_ptr[2];

      printf("*** Recovering at count %d, occupied = %ld,\n"
	     "*** areas = %p, sizes = %p, replay time = %.6f s\n",
	     misc_ptr->count, misc_ptr->occupied, areas, sizes,
	     (init_time.tv_sec-start_time.tv_sec)+
	     (init_time.tv_usec-start_time.tv_usec)/1e+6);

      assert(misc_ptr->occupied >= 0);

      /* In case the master process decided to be kind and wait for replay */
      if ( master_waits )
	kill(pid, SIGTSTP);	/* The master will wait for this and SIGCONT */

      for ( i = misc_ptr->count+1 ; i <= COUNT ; i++ ) {
	misc_ptr->count = i;
	slot = rand() % SLOTS;

	printf("Count:   %d  Slot: %d\n", i, slot);
	
	if ( areas[slot] ) {
	  if ( rand() % 1 ) {
	    lpsm_free(areas[slot]);
	    printf("Free:    %d (0x%08x) bytes at %p\n",
		   sizes[slot], sizes[slot], areas[slot]);
	    areas[slot] = NULL;
	    misc_ptr->occupied -= sizes[slot];
	    assert(misc_ptr->occupied >= 0);
	  } else {
	    rnd = (double)rand()/RAND_MAX;
	    newsize = (int)pow(2.0, MAXLOG*pow(rnd,3.0));
	    nptr = lpsm_realloc(areas[slot], newsize);
	    if ( nptr ) {
	      printf("Realloc: %d bytes at %p -> %d bytes at %p\n",
		     sizes[slot], areas[slot], newsize, nptr);
	      misc_ptr->occupied += newsize;
	      misc_ptr->occupied -= sizes[slot];
	      assert(misc_ptr->occupied >= 0);
	      sizes[slot] = newsize;
	      areas[slot] = nptr;
	      verify_no_overlap(slot);
	    } else {
	      printf("Realloc: %d bytes at %p -> %d bytes FAILED\n",
		     sizes[slot], areas[slot], newsize);
	    }
	  }
	} else {
	  rnd = (double)rand()/RAND_MAX;
	  sizes[slot] = (int)pow(2.0, MAXLOG*pow(rnd,5.0));
	  areas[slot] = lpsm_malloc(sizes[slot]);
	  if (areas[slot]) { 
	    printf("Alloc:   %d (0x%08x) bytes at %p\n",
		   sizes[slot], sizes[slot], areas[slot]);
	    misc_ptr->occupied += sizes[slot];
	  } else {
	    printf("Alloc:   %d (0x%08x) bytes FAILED\n",
		   sizes[slot], sizes[slot]);
	  }
	  assert(misc_ptr->occupied >= 0);
	  verify_no_overlap(slot);
	}
	
	if ( (i & 255) == 0 ) {
	  printf("Arena:   checkpointing, count = %d\n", i);
	  lpsm_checkpoint(0.1, PSMSYNC_SKIP);
	}
	printf("Arena:   %ld bytes officially occupied\n", misc_ptr->occupied);
      }
      /* If the loop finishes, exit... */
      fprintf(stderr, "ready...\n");
      lpsm_checkpoint(0.0, PSMSYNC_SYNC); /* Synchronous checkpoint */
      exit(0);
    }
  }

  return 0;
}
